

/************************************************************
*
* REXLANG - Using the PCA9685 16-Channel PWM/Servo Controller via I2C
* Written by Eric Novikoff
* Version 1.0 - 2020.04.21
* Adapted from Adafruit C++ and Waveshare Python drivers.
* Design Overview:
*    This driver accepts 16 inputs from the REXLANG block, each of which is a 
*    percent dutycycle for the corresponding PWM channel.  A 16-bit mask parameter
*    is used to determine which channels are active (default of zero is all channels on.)
*    A 16 bit invert mask inverts the output of channels individually since the the PCA9675
*    does not have this capability (default of zero is all channels uninverted.)
*    There is also a 'totempole' mask which controls whether the PCA9685 outputs are
*    totempole (push-pull) or open-drain.
*    The chip oscillator on thePCA9865 is approximate and there is a parameter to allow
*    tuning the driver to the actual chip frequency.  All channels of the PCA9865 operate
*    at the same output frequency, which is set by a parameter (defaults to 1Khz).
* Usage Notes:
*    It is recommended to run this driver in a separate task that operates at a longer
*    tick length than your project to avoid stretching a short tick time.  Setting 16 PWMs 
*    (one execution of the driver's 'main' function) takes about 34ms on a Raspberry Pi
*    3B+, determined largely by the default I2C bus clock speed of 62KHz, which could be
*    altered with external software such as the gpio utility.
*
*************************************************************/

/*
 *       PCA 9685 Device Constants
 */
 
// REGISTER ADDRESSES
#define PCA9685_MODE1 0x00      /**< Mode Register 1 */
#define PCA9685_MODE2 0x01      /**< Mode Register 2 */
#define PCA9685_SUBADR1 0x02    /**< I2C-bus subaddress 1 */
#define PCA9685_SUBADR2 0x03    /**< I2C-bus subaddress 2 */
#define PCA9685_SUBADR3 0x04    /**< I2C-bus subaddress 3 */
#define PCA9685_ALLCALLADR 0x05 /**< LED All Call I2C-bus address */
#define PCA9685_LED0_ON_L 0x06  /**< LED0 on tick, low byte*/
#define PCA9685_LED0_ON_H 0x07  /**< LED0 on tick, high byte*/
#define PCA9685_LED0_OFF_L 0x08 /**< LED0 off tick, low byte */
#define PCA9685_LED0_OFF_H 0x09 /**< LED0 off tick, high byte */
// etc all 16:  LED15_OFF_H 0x45
#define PCA9685_ALLLED_ON_L 0xFA  /**< load all the LEDn_ON registers, low */
#define PCA9685_ALLLED_ON_H 0xFB  /**< load all the LEDn_ON registers, high */
#define PCA9685_ALLLED_OFF_L 0xFC /**< load all the LEDn_OFF registers, low */
#define PCA9685_ALLLED_OFF_H 0xFD /**< load all the LEDn_OFF registers,high */
#define PCA9685_PRESCALE 0xFE     /**< Prescaler for PWM output frequency */
#define PCA9685_TESTMODE 0xFF     /**< defines the test mode to be entered */

// MODE1 bits
#define MODE1_ALLCAL 0x01  /**< respond to LED All Call I2C-bus address */
#define MODE1_SUB3 0x02    /**< respond to I2C-bus subaddress 3 */
#define MODE1_SUB2 0x04    /**< respond to I2C-bus subaddress 2 */
#define MODE1_SUB1 0x08    /**< respond to I2C-bus subaddress 1 */
#define MODE1_SLEEP 0x10   /**< Low power mode. Oscillator off */
#define MODE1_AI 0x20      /**< Auto-Increment enabled */
#define MODE1_EXTCLK 0x40  /**< Use EXTCLK pin clock */
#define MODE1_RESTART 0x80 /**< Restart enabled */

// MODE2 bits
#define MODE2_OUTNE_0 0x01 /**< Active LOW output enable input */
#define MODE2_OUTNE_1 0x02 /**< Active LOW output enable input - high impedience */
#define MODE2_OUTDRV 0x04 /**< totem pole structure vs open-drain */
#define MODE2_OCH 0x08    /**< Outputs change on ACK vs STOP */
#define MODE2_INVRT 0x10  /**< Output logic state inverted */

#define PCA9685_PRESCALE_MIN 3   /**< minimum prescale value */
#define PCA9685_PRESCALE_MAX 255 /**< maximum prescale value */

#define PCA9685_I2C_ADDRESS 0x40      /**< Default PCA9685 I2C Slave Address */
#define FREQUENCY_OSCILLATOR 25000000 /**< Int. osc. frequency in datasheet */
  /*
   * In theory the internal oscillator (clock) is 25MHz but it really isn't
   * that precise. You can 'calibrate' this by tweaking this number until
   * you get the PWM update frequency you're expecting!
   * The int.osc. for the PCA9685 chip is a range between about 23-27MHz and
   * is used for calculating things like writeMicroseconds()
   * Analog servos run at ~50 Hz updates, It is importaint to use an
   * oscilloscope in setting the int.osc frequency for the I2C PCA9685 chip.
   * 1) Attach the oscilloscope to one of the PWM signal pins and ground on
   *    the I2C PCA9685 chip you are setting the value for.
   * 2) Adjust setOscillatorFrequency() until the PWM update frequency is the
   *    expected value (50Hz for most ESCs)
   * Setting the value here is specific to each individual I2C PCA9685 chip and
   * affects the calculations for the PWM update frequency. 
   * Failure to correctly set the int.osc value will cause unexpected PWM results
   */

/*
 * REXLANG Parameters
 */
string parameter(0) i2c_dev; // the I2C bus is defined by the p0 parameter
long parameter(1) i2c_addr; // Overrides the default setting above if nonzero.  Note this is a DECIMAL parameter!
long parameter(2) pwmfrequency; // Frequency of PWM output, must be less than 1500 (Hz)
long parameter(3) oscillatorfreq; // Measured input oscillator frequency, overrides default above if nonzero (in Hz)
long parameter(4) totempole; // enable all outputs as totem pole if 1, enable open-drain outputs if 0
long parameter(5) activemask; //16-bit value where each active bit enables the corresponding output [pin15]...[pin0]
long parameter(6) invertmask; //16-bit value where each active bit inverts the pwem on the corresponding output [pin15]...[pin0]

/*
 * REXLANG Input Pins - assigned to read only variables
 */
long input(0) dutycycle0; // Signal controlling the percent duty cycle of output, 0 <= dutycycle <= 100
long input(1) dutycycle1;
long input(2) dutycycle2;
long input(3) dutycycle3;
long input(4) dutycycle4;
long input(5) dutycycle5;
long input(6) dutycycle6;
long input(7) dutycycle7;
long input(8) dutycycle8;
long input(9) dutycycle9;
long input(10) dutycycle10;
long input(11) dutycycle11;
long input(12) dutycycle12;
long input(13) dutycycle13;
long input(14) dutycycle14;
long input(15) dutycycle15;


//declaration of variables
long i2c_bufTx[5]; //buffer for transmitting data
long i2c_bufRx[5]; //buffer for receiving data
long i2c_bus_handle;
long i2c_chip_address;
long i2c_write_count;
long i2c_read_count;
long i2c_ret_fun;

long read8(long addr)
{
    Trace(1, "read");
    Trace(1, addr);
    
    i2c_bufTx[0] = addr; //register no.
    i2c_write_count = 1;
    i2c_read_count = 1;
    i2c_ret_fun = I2C(i2c_bus_handle, i2c_chip_address, i2c_bufTx, i2c_write_count, i2c_bufRx, i2c_read_count);
    Trace(1, i2c_bufRx[0]);
    return i2c_bufRx[0] & 0xff;
}
    
void write8(long addr, long data)
{
    Trace(2, "write");
    Trace(2, addr);
    Trace(2, data);
    
    i2c_bufTx[0] = addr & 0xFF; //register no. 
    i2c_bufTx[1] = data & 0xFF;
    i2c_write_count = 2;
    i2c_read_count = 0;
    i2c_ret_fun = I2C(i2c_bus_handle, i2c_chip_address, i2c_bufTx, i2c_write_count, i2c_bufRx, i2c_read_count);
}

/*
 *  Sets the PWM output of one of the PCA9685 pins
 *  num: One of the PWM output pins, from 0 to 15
 *  on: At what point in the 4096-part cycle to turn the PWM output ON
 *  off: At what point in the 4096-part cycle to turn the PWM output OFF
 */
void setPWM(long chan, long onper, long offper)
{
    Trace(3, "setPWM");
    Trace(3, chan);
    Trace(3, onper);
    Trace(3, offper); 
    
    i2c_bufTx[0] = PCA9685_LED0_ON_L + 4 * chan;
    i2c_bufTx[1] = onper & 0x00FF;
    i2c_bufTx[2] = (onper >> 8) & 0x00FF;
    i2c_bufTx[3] = offper & 0x00FF;
    i2c_bufTx[4] = (offper >> 8) & 0x00FF;
    i2c_write_count = 5;
    i2c_read_count = 0;
    i2c_ret_fun = I2C(i2c_bus_handle, i2c_chip_address, i2c_bufTx, i2c_write_count, i2c_bufRx, i2c_read_count);
}

/*
 * Helper to set a pin's PWM output.  Sets pin without having to deal without
 * on/off tick placement and properly handles a zero value as completely off and
 * 4095 as completely on.  Optional invert parameter supports inverting the pulse
 * for sinking a current to ground.
 * num: One of the PWM output pins, from 0 to 15
 * val: The number of ticke out of 4096 to be active, should be a value from 0 to
 *      4095, inclusive.
 * invert: invert output if true (nonzero)
 */
void setPin(long num, long val, long invert)
{
    Trace(4, "SetPin");
    Trace(4,num);
    Trace(4,val);
    Trace(4,invert);
    
    // Clamp value between 0 and 4095 inclusive
    val = (long)max(min((double)val, 4095.0), 0.0);

    if (invert) {
        if (val == 0) {
            // Special value for signal fully on.
            setPWM(num, 4096, 0);
        } else if (val == 4095) {
            // Special value for signal fully off.
            setPWM(num, 0, 4096);
        } else {
            setPWM(num, 0, 4095 - val);
        }
    } else {
        if (val == 4095) {
            // Special value for signal fully on.
            setPWM(num, 4096, 0);
        } else if (val == 0) {
            // Special value for signal fully off.
            setPWM(num, 0, 4096);
        } else {
            setPWM(num, 0, val);
        }
    }
}

//the init procedure is executed once when the REXLANG function block initializes
long init(void)
{
    long freq = pwmfrequency;
    long oscfreq = oscillatorfreq;
    double prescaleval;
    long prescale;
    long oldmode;
    long newmode;

    Trace(0,"init");
    
    // Setup I2C
    i2c_bus_handle = OpenI2C(i2c_dev); // open I2C bus
    if (i2c_addr) i2c_chip_address = i2c_addr; // 7-bit address of the I2C device
    else i2c_chip_address = PCA9685_I2C_ADDRESS;
    
    // Initialize the PCA9685
    // write8(PCA9685_MODE1, MODE1_RESTART); - Adafruit approach
    write8(PCA9685_MODE1, 0x00);
 
    
    // Set the frequency for the entire chip's channels, up to ~1.6KHz
    // Attempts to match the frequency parameter as much as possible
    // External clock option not supported with this release
    if (!freq) freq = 1000;
    if (freq < 1) freq = 1; else if (freq > 3500) freq = 3500; // Datasheet limit is 3052=50MHz/(4*4096)
    // Calculate prescale value
    if (!oscfreq) oscfreq = FREQUENCY_OSCILLATOR;
    prescaleval = ((oscfreq / (freq * 4096.0)) + 0.5) - 1;
    if (prescaleval < PCA9685_PRESCALE_MIN) prescaleval = PCA9685_PRESCALE_MIN;
    if (prescaleval > PCA9685_PRESCALE_MAX) prescaleval = PCA9685_PRESCALE_MAX;
    prescale = (long)prescaleval;
    oldmode = read8(PCA9685_MODE1);
    newmode = oldmode | MODE1_SLEEP;
    write8(PCA9685_MODE1, newmode); //go to sleep
    write8(PCA9685_PRESCALE, prescale); //set prescaleval
    Sleep(0.01); // allow oscillator to stabilize
    
    // Set up totempole (push-pull)or open drain outputs
    oldmode = read8(PCA9685_MODE2);
    if (totempole) { newmode = oldmode | MODE2_OUTDRV; } else { newmode = oldmode & ~MODE2_OUTDRV; }
    write8(PCA9685_MODE2, newmode);
    Sleep(0.01);
    
    //Set the MODE1 register to turn on auto increment and wake up the PCA9685
    write8(PCA9685_MODE1, oldmode & ~MODE1_SLEEP);
    Sleep(0.01);
    write8(PCA9685_MODE1, oldmode | MODE1_AI | MODE1_RESTART);
    
    return 0;
}

//the main procedure is executed once in each sampling period
long main(void)
{
    long am;
    long im;
    
    Trace(1,"main");
    
    if (!activemask) am = 0xFFFF; else am = activemask;
    if (!invertmask) im = 0x0000; else im = invertmask;
    
    if (am & 0x0001) setPin(0, (long)(dutycycle0*40.95), im & 0x0001);
    if (am & 0x0002) setPin(1, (long)(dutycycle1*40.95), im & 0x0002);
    if (am & 0x0004) setPin(2, (long)(dutycycle2*40.95), im & 0x0004);
    if (am & 0x0008) setPin(3, (long)(dutycycle3*40.95), im & 0x0008);
    if (am & 0x0010) setPin(4, (long)(dutycycle4*40.95), im & 0x0010);
    if (am & 0x0020) setPin(5, (long)(dutycycle5*40.95), im & 0x0020);
    if (am & 0x0040) setPin(6, (long)(dutycycle6*40.95), im & 0x0040);
    if (am & 0x0080) setPin(7, (long)(dutycycle7*40.95), im & 0x0080);
    if (am & 0x0100) setPin(8, (long)(dutycycle8*40.95), im & 0x0100);
    if (am & 0x0200) setPin(9, (long)(dutycycle9*40.95), im & 0x0200);
    if (am & 0x0400) setPin(10, (long)(dutycycle10*40.95), im & 0x0400);
    if (am & 0x0800) setPin(11, (long)(dutycycle11*40.95), im & 0x0800);
    if (am & 0x1000) setPin(12, (long)(dutycycle12*40.95), im & 0x1000);
    if (am & 0x2000) setPin(13, (long)(dutycycle13*40.95), im & 0x2000);
    if (am & 0x4000) setPin(14, (long)(dutycycle14*40.95), im & 0x4000);
    if (am & 0x8000) setPin(15, (long)(dutycycle15*40.95), im & 0x8000);
    
    return 0;
}

//the exit procedure is executed once when the task is correctly terminated
// (system shutdown, downloading new control algorithm, etc.)
long exit(void)
{
    return 0;
}
